// DRGTexture.cpp
// ------------------------------------------------------------------------------------
// Copyright  1999 Intel Corporation
// All Rights Reserved
// 
// Permission is granted to use, copy, distribute and prepare derivative works of this 
// software for any purpose and without fee, provided, that the above copyright notice
// and this statement appear in all copies.  Intel makes no representations about the
// suitability of this software for any purpose.  This software is provided "AS IS." 
//
// Intel specifically disclaims all warranties, express or implied, and all liability,
// including consequential and other indirect damages, for the use of this software, 
// including liability for infringement of any proprietary rights, and including the 
// warranties of merchantability and fitness for a particular purpose.  Intel does not
// assume any responsibility for any errors which may appear in this software nor any
// responsibility to update it.
// ------------------------------------------------------------------------------------
//
//  PURPOSE:
//    
// DRGTexture.cpp: implementation of the CDRGTexture class.
// Originally adapted from some DX6 sample code. Added some robustness and extra 
// functionality (alpha bitmaps, creating inverse hieghtmaps for embossing, automatic
// resize for HW requirements, etc)
//
// ------------------------------------------------------------------------------------
//
// Authors: Dean Macri, Kim Pallister - Intel Developer Relations & Engineering Group
//
////////////////////////////////////////////////////////////////////////////////////////

#include <tchar.h>

#include "DRGShell.h"

struct TEXTURESEARCHINFO
{
	DWORD dwDesiredBPP; // Input for texture format search
	BOOL  bUseAlpha;
	DWORD dwDesiredAlphaBPP;
	BOOL  bUsePalette;
	BOOL  bUseFourCC;
	BOOL  bUseBumpDUDV;	
	BOOL  bFoundGoodFormat;

	DDPIXELFORMAT* pddpf; // Result of texture format search
};

inline BOOL FileExists( TCHAR* n )
{ FILE* fp=_tfopen(n,TEXT("rb")); return fp?(0==fclose(fp)):FALSE; }


//-----------------------------------------------------------------------------
// Name: TextureSearchCallback()
// Desc: Enumeration callback routine to find a best-matching texture format. 
//       The param data is the DDPIXELFORMAT of the best-so-far matching
//       texture. Note: the desired BPP is passed in the dwSize field, and the
//       default BPP is passed in the dwFlags field.
// Return: If the current texture format being looked at is a good match, we set
//       the ptsi->bFoundGoodFormat to true. Otherwise, it remains false.
//		 either way, returning DDENUMRET_OK indicates to continue the enumeration
//       process (which we do when we realize the current format is no good). 
//		 Returning DDENUMRET_CANCEL indicates to stop enumerations (which we want
//		 to do if we have found a good texture format).
//-----------------------------------------------------------------------------
static HRESULT CALLBACK TextureSearchCallback( DDPIXELFORMAT* pddpf, VOID* param )
{
	//if bad data in...
    if( NULL==pddpf || NULL==param )
        return DDENUMRET_OK;

	TEXTURESEARCHINFO* ptsi = (TEXTURESEARCHINFO*)param;

    if (ptsi->bUseBumpDUDV)
	{
		//for now, I haven't supported the luminance flags, that's next
		if ( pddpf->dwFlags & (DDPF_BUMPLUMINANCE | DDPF_LUMINANCE))
		{
			//this tex format is not the bump map we want
			return DDENUMRET_OK;
		}
		// make sure we get a BUMPDUDV mode;		
		if (!( pddpf->dwFlags & DDPF_BUMPDUDV ))
		{
			//this tex format is not a bump map
			return DDENUMRET_OK;
		}
		else
		{
			OutputDebugString("desired bump map format found \n");
		}
	}
	else
	{
		// Skip any bump mapping modes, since they werem't requested 
		if( pddpf->dwFlags & (DDPF_LUMINANCE|DDPF_BUMPLUMINANCE|DDPF_BUMPDUDV) )
			return DDENUMRET_OK;	
	}

	// Check for palettized formats
	if( ptsi->bUsePalette )
	{
		if( !( pddpf->dwFlags & DDPF_PALETTEINDEXED8 ) )
			return DDENUMRET_OK;

		// Accept the first 8-bit palettized format we get
        memcpy( ptsi->pddpf, pddpf, sizeof(DDPIXELFORMAT) );
		ptsi->bFoundGoodFormat = TRUE;
        return DDENUMRET_CANCEL;
    }

	// Else, skip any paletized formats (all modes under 16bpp)
	// NOTE, THAT MEANS WE ARE NOT CURRENTLY HANDLING PALETTIZED TEXTURES
	if( pddpf->dwRGBBitCount < 16 )
		return DDENUMRET_OK;

	// Check for FourCC formats
    if( ptsi->bUseFourCC )
	{
		if( pddpf->dwFourCC == 0 )
		    return DDENUMRET_OK;

		return DDENUMRET_CANCEL;
	}
	// Else, skip any FourCC formats
	if( pddpf->dwFourCC != 0 )
		return DDENUMRET_OK;

	// Make sure current alpha format agrees with requested format type
	// (i.e. If the user asks for alpha, and the surface doesn't have an
	// alpha channel, or vice versa, then we want to get out)
	if( (ptsi->bUseAlpha==TRUE) && !(pddpf->dwFlags&DDPF_ALPHAPIXELS) )
		return DDENUMRET_OK;
	if( (ptsi->bUseAlpha==FALSE) && (pddpf->dwFlags&DDPF_ALPHAPIXELS) )
		return DDENUMRET_OK;

    // Check if we found a good match
    if( pddpf->dwRGBBitCount == ptsi->dwDesiredBPP )
    {
		//if requesting to use alpha, then format must have required alpha
		if ((ptsi->bUseAlpha) || (ptsi->dwDesiredAlphaBPP == 1))
		{
			// dwAlphaBitDepth member reflects the bit depth of an alpha-only 
			// pixel format. For pixel formats with RGBA, we have to 
			// do the following
			
			if ((pddpf->dwFlags && DDPF_ALPHAPIXELS)==0)				
			{
				//rejecting, format had no alpha
				return	DDENUMRET_OK;		
			}
			else
			{
				if	((ptsi->dwDesiredAlphaBPP == 4) && (pddpf->dwRGBAlphaBitMask != 0xf000))
				{
					//rejecting. Format had alpha, but not 4 bits
					return	DDENUMRET_OK;		
				}
				if	((ptsi->dwDesiredAlphaBPP == 8) && (pddpf->dwRGBAlphaBitMask != 0xff000000))
				{
					//rejecting. Format had alpha, but not 8 bits
					return	DDENUMRET_OK;		
				}
			}					
			
			//else successful, so fall through
		}

		//if we fell through to here, then we found what we wanted
        memcpy( ptsi->pddpf, pddpf, sizeof(DDPIXELFORMAT) );
		ptsi->bFoundGoodFormat = TRUE;
        return DDENUMRET_CANCEL;
    }

	//if we fell through to here, then the bpp didn't match what we wanted
    return DDENUMRET_OK;
}


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDRGTexture::CDRGTexture()
{
	//init the member variables.
	hbmBitmap = NULL;
	hbmAlphaBitmap = NULL;
    pddsSurface = NULL;
    ptexTexture = NULL;
	lstrcpy( strName, "" ); 
	lstrcpy( strAlphaName, "" );

    dwStage = 0;	
    bHasAlpha = false;
	dwFlags = 0;	
}

CDRGTexture::~CDRGTexture()
{
	//init the member variables.
	if (hbmBitmap)
		hbmBitmap = NULL;
	
	if (hbmAlphaBitmap)		
		hbmAlphaBitmap = NULL;
    
    if (pddsSurface)
	{
		pddsSurface->Release();
		pddsSurface = NULL;
	}

    dwStage = 0;
	bHasAlpha = false;
	dwFlags = 0;
}


//-----------------------------------------------------------------------------
// Name: D3DTextr_CreateTexture()
// Desc: Is passed a filename and creates a local Bitmap from that file.
//       The texture can not be used until it is restored, however.
//-----------------------------------------------------------------------------
HRESULT CDRGTexture::CreateTexture( TCHAR* strTextureName, DWORD dwTextureStage, DWORD dwTextureFlags )
{

	//copy parameters into member vars
    lstrcpy( strName, strTextureName );
	lstrcpy( strAlphaName, "" ); //since the user is not going to use a bitmap for alpha bits
    dwStage = dwTextureStage;
    dwFlags = dwTextureFlags;

    // Create a bitmap and load the texture file into it,
    if( FAILED( LoadTextureImage() ) )
        return E_FAIL;

    return S_OK;
}

HRESULT CDRGTexture::CreateTexture( TCHAR* strTextureName, TCHAR* strAlphaTextureName, DWORD dwTextureStage, DWORD dwTextureFlags )
{

	//copy parameters into member vars
    lstrcpy( strName, strTextureName );
	lstrcpy( strAlphaName, strAlphaTextureName );
    dwStage = dwTextureStage;
    dwFlags = dwTextureFlags;

	//because this function got called, we are trying to load an alpha texture as well
	bHasAlpha = true;

    // Create a bitmap and load the texture file into it,
    if( FAILED( LoadTextureImage() ) )
        return E_FAIL;

    return S_OK;
}


HRESULT CDRGTexture::CreateTexture(  CDRGWrapper *pWrapper, TCHAR*	strTextureName, DWORD dwHres, DWORD dwVres,
									DWORD dwBpp, DWORD dwTextureStage, DWORD dwTextureFlags, DWORD	*pBits)
{
	lstrcpy( strName, strTextureName );
	lstrcpy( strAlphaName, "" ); //since the user is not going to use a bitmap for alpha bits
    dwStage = dwTextureStage;
    dwFlags = dwTextureFlags;

	// CREATE A TEMP SURFACE IN SYSMEM WITH SIMILAR FORMAT AS THE DEVICE
	// (we're just using this to get a DC)
	
	// Setup the new surface desc
    DDSURFACEDESC2 ddsd;
	ZeroMemory( &ddsd, sizeof(DDSURFACEDESC2) );
    ddsd.dwSize                 = sizeof(DDSURFACEDESC2);
    ddsd.dwFlags                = 0;
    ddsd.ddsCaps.dwCaps         = 0;
    ddsd.ddpfPixelFormat.dwSize = sizeof(DDPIXELFORMAT);
	
	// Get info on the target surface, so we can create one similar, but in sys mem
    pWrapper->m_pBackBuffer->GetSurfaceDesc( &ddsd );
    ddsd.dwFlags          = DDSD_CAPS|DDSD_HEIGHT|DDSD_WIDTH|DDSD_PIXELFORMAT|
		                    DDSD_TEXTURESTAGE;
    ddsd.ddsCaps.dwCaps   = DDSCAPS_TEXTURE|DDSCAPS_SYSTEMMEMORY;
    ddsd.ddsCaps.dwCaps2  = 0L;
    ddsd.dwWidth          = 16; // Since we are only using this to get a compatible DC, 
    ddsd.dwHeight         = 16; // there is no need to create anything very large

    // Create the temp sysmem surface
    LPDIRECTDRAWSURFACE4 pddsTempSurface;
	HRESULT hr;
    if( FAILED( hr = pWrapper->m_pDD->CreateSurface( &ddsd, &pddsTempSurface, NULL ) ) )
	{
        OutputDebugString("Couldn't create a temp surface \n");
		return -1;
	}

	HDC hdcBitmap; 

	//get a DC from the temp surface
	hr = pddsTempSurface->GetDC(&hdcBitmap);
	if (hr == DD_OK)
	{
		hbmBitmap = CreateCompatibleBitmap(  hdcBitmap, dwHres,  dwVres);
		hr = pWrapper->m_pFrontBuffer->ReleaseDC(hdcBitmap);
		pddsTempSurface->Release();
	}

	// If array of bits was supplied to fill bitmap, copy them over
	if(pBits)
	{
		if (SetBitmapBits( hbmBitmap, (4* dwHres * dwVres), pBits) == 0)
			return -1;
	}  
		
	return S_OK;    
}



//-----------------------------------------------------------------------------
// Name: LoadTextureImage()
// Desc: Loads a texture map file (or resource) into a BITMAP surface.
//-----------------------------------------------------------------------------
HRESULT CDRGTexture::LoadTextureImage()
{
	TCHAR* strFilename = &strName[0];
    TCHAR* strExtension;

	TCHAR* strAlphaFilename = &strAlphaName[0];

        // Get the filename extension
    if( NULL == ( strExtension = _tcsrchr( strFilename, TEXT('.') ) ) )
        return DDERR_UNSUPPORTED;

	// The load from resource is part of the MS sample code we started with. Haven't tested it.
    // Check the executable's resource, and return if found.
    if( NULL != ( hbmBitmap = (HBITMAP)LoadImage( GetModuleHandle(NULL),strFilename, 
												  IMAGE_BITMAP, 0, 0, 
												  LR_CREATEDIBSECTION) ) )
	{
		if (bHasAlpha)
		{
			//if the bool says there is alpha, then it was either that a bitmap for alpha,
			//or the user requested a 1 bit color key
			if (strAlphaFilename != "")
			{
				if( NULL != ( hbmAlphaBitmap = (HBITMAP)LoadImage( GetModuleHandle(NULL),strAlphaFilename, 
																  IMAGE_BITMAP, 0, 0, 
																  LR_CREATEDIBSECTION) ) )
					return DD_OK;
				else 
				{
					//really, we should look elsewhere, but for now, I will disable alpha.
					bHasAlpha = false;					
				}
			}
			else
			{
				// alpha bitmap null, but user requested alpha.
				hbmAlphaBitmap = NULL;
				return DD_OK;
			}
		}
		else
		{
			hbmAlphaBitmap = NULL;
		}
        return DD_OK;
	}
	        
	
	//else, check files
    if( !lstrcmpi( strExtension, ".bmp" ) )
    {

        // If the bitmap wasn't a resource, try it as a file.
        if( NULL == hbmBitmap )
		{
			if (NULL == (hbmBitmap = (HBITMAP)LoadImage( NULL, strFilename, IMAGE_BITMAP, 0, 0, 
													LR_LOADFROMFILE|LR_CREATEDIBSECTION ) ))
				return DDERR_NOTFOUND;
		}

		//If we get here, we have found the bitmap and loaded it, 
		
		//Now look for the alpha bitmap
		if (bHasAlpha)
		{
			if ((NULL == hbmAlphaBitmap) && (NULL != strAlphaFilename))
			{
				//if they asked for an alpha bitmap, but we couldn't find one, we will just disable alpha
				//we should probably fail instead though!
				if (NULL == (hbmAlphaBitmap = (HBITMAP)LoadImage( NULL, strAlphaFilename, 
													 IMAGE_BITMAP, 0, 0, 
													 LR_LOADFROMFILE|LR_CREATEDIBSECTION) ))
					return DDERR_NOTFOUND;									 
			}
		}

		//last sanity check.
		if (hbmBitmap)
		{
			if (bHasAlpha)
				if (hbmAlphaBitmap)
					return DD_OK; // alpha and bitmap found
				else
					return DDERR_NOTFOUND; //asked for alpha, but alpha not found
			else
				return DD_OK; //no alpha, bitmap found
		}
		else
		{
			return DDERR_NOTFOUND; //bitmap not found
		}
    }

	// Could check for other file formats here (jpg, png, etc)(an EFTU) 
	
    return DDERR_UNSUPPORTED;
}


//-----------------------------------------------------------------------------
// Name: FindTextureFile()
// Desc: Looks for the specified file in the current directory, or in the path
//       specified in the environment variables. After finding it, the function
//       returns the full path.
//-----------------------------------------------------------------------------
HRESULT CDRGTexture::FindTextureFile( TCHAR* strFilename, TCHAR* strTexturePath,
                         TCHAR* strFullPath )
{
    // First, check the current directory
    _tcscpy( strFullPath, strFilename );
    if( FileExists( strFullPath ) )
        return DD_OK;

    // Next, check to see if an environment variable specifies the path
    TCHAR* strPath = _tgetenv( TEXT("DRGPATH") );
    if( NULL != strPath )
    {
        _stprintf( strFullPath, TEXT("%s\\%s"), strPath, strFilename );
        if( FileExists( strFullPath ) )
            return DD_OK;
    }

    return DDERR_NOTFOUND;
}


//-----------------------------------------------------------------------------
// Name: DRGTexturer:Restore()
// Desc: Invalidates the current texture object and rebuilds it with the 
//       given device
//-----------------------------------------------------------------------------
HRESULT CDRGTexture::Restore(LPDIRECT3DDEVICE3 pd3dDevice )
{
    // Check params
    if( NULL == pd3dDevice )
        return DDERR_INVALIDPARAMS;

	// Release any previously created objects
    SAFE_RELEASE( ptexTexture );
    SAFE_RELEASE( pddsSurface );

	// Restore the texture surface from the bitmap image.
	return RestoreFromBitmap( pd3dDevice );
}


//-----------------------------------------------------------------------------
// Name: RestoreFromBitmap()
// Desc: Invalidates the current texture objects and rebuilds new ones
//       using the new device.
//-----------------------------------------------------------------------------
HRESULT CDRGTexture::RestoreFromBitmap( LPDIRECT3DDEVICE3 pd3dDevice )
{
    // Get the DirectDraw interface for creating surfaces
    LPDIRECTDRAW4 pDD;
    if( NULL == ( pDD = GetDirectDrawFromDevice( pd3dDevice ) ) )
        return E_FAIL;
    pDD->Release();

    // Get the device caps
    D3DDEVICEDESC ddHwDesc, ddSwDesc;
    DWORD         dwDeviceCaps;
    ddHwDesc.dwSize = sizeof(D3DDEVICEDESC);
    ddSwDesc.dwSize = sizeof(D3DDEVICEDESC);
    if( FAILED( pd3dDevice->GetCaps( &ddHwDesc, &ddSwDesc ) ) )
        return E_FAIL;
    
	bool bHardware = false;

    if( ddHwDesc.dwFlags != 0x00000000) 
		bHardware = true;

	if (bHardware == true)
		dwDeviceCaps = ddHwDesc.dpcTriCaps.dwTextureCaps;
    else
		dwDeviceCaps = ddSwDesc.dpcTriCaps.dwTextureCaps;

	// Get the bitmap structure (to extract width, height, and bpp)
    BITMAP bm;
    //HBITMAP hbmBitmap = ptcTexture->hbmBitmap;
    GetObject( hbmBitmap, sizeof(BITMAP), &bm );
    DWORD dwWidth  = (DWORD)bm.bmWidth;
    DWORD dwHeight = (DWORD)bm.bmHeight;

	//WARNING. WE SHOULD DO SOMETHING HERE IF THE ALPHA TEXTURE IS A DIFFERENT SIZE!
	//should at least fail gracefully. Ideally, we could do something that scales the
	//alpha bitmap to fit the RGB bitmap

    // Setup the new surface desc
    DDSURFACEDESC2 ddsd;
    ZeroMemory( &ddsd, sizeof(DDSURFACEDESC2) );
    ddsd.dwSize                 = sizeof(DDSURFACEDESC2);
    ddsd.dwFlags                = 0;
    ddsd.ddsCaps.dwCaps         = 0;
    ddsd.ddpfPixelFormat.dwSize = sizeof(DDPIXELFORMAT);


    ddsd.dwFlags         = DDSD_CAPS|DDSD_HEIGHT|DDSD_WIDTH|
                           DDSD_PIXELFORMAT|DDSD_TEXTURESTAGE;//|DDSD_ALPHABITDEPTH;
    
	//handle case where the application specified it would want to render to it
    if (dwFlags & DRGTEXTR_RENDERTARGET)
	{
		if (dwDeviceCaps & D3DDEVCAPS_SEPARATETEXTUREMEMORIES)
		{
			//should handle somehow, maybe blitting and the like
			OutputDebugString("WARNING! DEV HAS SEPARATE TEXTURE MEMORIES. CREATING TARGET TEXTURE LIKELY WONT WORK\n");
		}

		if (bHardware)
			ddsd.ddsCaps.dwCaps  = (DDSCAPS_TEXTURE | DDSCAPS_3DDEVICE |  DDSCAPS_VIDEOMEMORY);
		else
			ddsd.ddsCaps.dwCaps  = (DDSCAPS_TEXTURE | DDSCAPS_3DDEVICE |  DDSCAPS_SYSTEMMEMORY);

		//we can't use DX6 texture management if we are doing target textures
		//ddsd.ddsCaps.dwCaps2 = 0;//DDSCAPS2_TEXTUREMANAGE;
	}
	else
	{
		ddsd.ddsCaps.dwCaps  = DDSCAPS_TEXTURE;
		ddsd.ddsCaps.dwCaps2 = DDSCAPS2_TEXTUREMANAGE;
	}

	//now handle creation of a bump map
	BOOL bUseBumpDUDV = false;
	if (dwFlags & DRGTEXTR_CREATEBUMPMAP_DUDV)
	{
		bUseBumpDUDV = true;
		ddsd.ddpfPixelFormat.dwFlags = (ddsd.ddpfPixelFormat.dwFlags | DDPF_BUMPDUDV);
		ddsd.ddpfPixelFormat.dwBumpBitCount          = 16;
		ddsd.ddpfPixelFormat.dwBumpDuBitMask         = 0x000000ff;
		ddsd.ddpfPixelFormat.dwBumpDvBitMask         = 0x0000ff00;
		ddsd.ddpfPixelFormat.dwBumpLuminanceBitMask  = 0x00000000;

		if (bHardware)
			ddsd.ddsCaps.dwCaps  |= DDSCAPS_VIDEOMEMORY;
		else
			ddsd.ddsCaps.dwCaps  |= DDSCAPS_SYSTEMMEMORY;

		ddsd.ddsCaps.dwCaps2 = 0;//DDSCAPS2_TEXTUREMANAGE;
	}

	
	//caution: The dwStage member doesn't matter on a lot of HW, but needs to be set explicitly
	// on HW where textures are stored in separate memory spaces depending on stage (3Dfx 
	// Voodoo 2)for example)
    ddsd.dwTextureStage  = dwStage;
    ddsd.dwWidth         = dwWidth;
    ddsd.dwHeight        = dwHeight;
	
    // Adjust width and height, if the driver requires it...
	
	//   for 'requires power of 2'
    if( dwDeviceCaps & D3DPTEXTURECAPS_POW2 )
    {
        for( ddsd.dwWidth=1;  dwWidth>ddsd.dwWidth;   ddsd.dwWidth<<=1 );
        for( ddsd.dwHeight=1; dwHeight>ddsd.dwHeight; ddsd.dwHeight<<=1 );
    }
	
	//   for 'requires square texture'
    if( dwDeviceCaps & D3DPTEXTURECAPS_SQUAREONLY )
    {
        if( ddsd.dwWidth > ddsd.dwHeight ) ddsd.dwHeight = ddsd.dwWidth;
        else                               ddsd.dwWidth  = ddsd.dwHeight;
    }
	
	//   for 'I tried to create a texture bigger than the HW supports and I can't get up'
	if(bHardware) 
	{
		if ((ddHwDesc.dwMaxTextureWidth < dwWidth) || (ddHwDesc.dwMaxTextureHeight < dwHeight))
		{	
			while(ddHwDesc.dwMaxTextureWidth < dwWidth)
				dwWidth>>=1;
			while(ddHwDesc.dwMaxTextureHeight < dwHeight)
				dwHeight>>=1;
			ddsd.dwWidth         = dwWidth;
			ddsd.dwHeight        = dwHeight;
		}
	}
    else
	{
		if ((ddSwDesc.dwMaxTextureWidth < dwWidth) || (ddSwDesc.dwMaxTextureHeight < dwHeight))
		{
			while(ddSwDesc.dwMaxTextureWidth < dwWidth)
				dwWidth>>=1;
			while(ddSwDesc.dwMaxTextureHeight < dwHeight)
				dwHeight>>=1;
			ddsd.dwWidth         = dwWidth;
			ddsd.dwHeight        = dwHeight;
		}
	}

	BOOL bUsePalette = ( bm.bmBitsPixel <= 8 );
	BOOL bUseAlpha   = FALSE;
		
	if( bHasAlpha )
		bUseAlpha = bHasAlpha;
	
	if( dwFlags & (DRGTEXTR_TRANSPARENTWHITE|DRGTEXTR_TRANSPARENTBLACK) )
	{
		if( bUsePalette )
		{
			if( dwDeviceCaps & D3DPTEXTURECAPS_ALPHAPALETTE )
			{
				bUseAlpha   = TRUE; 
				bUsePalette = TRUE; 
			}
		}
		else
		{
			bUseAlpha   = TRUE; 
			bUsePalette = FALSE;
		}
	}
	
	// Setup the structure to be used for texture enumeration.
	TEXTURESEARCHINFO tsi;
	tsi.pddpf            = &ddsd.ddpfPixelFormat;
	tsi.bUseAlpha        = bUseAlpha;
	tsi.bUsePalette      = bUsePalette;
	tsi.bUseFourCC       = ( ddsd.ddpfPixelFormat.dwFlags & DDPF_FOURCC );
	tsi.bUseBumpDUDV	 = bUseBumpDUDV;
	tsi.dwDesiredBPP     = 16;
	tsi.bFoundGoodFormat = FALSE;
	if (bHasAlpha)
		tsi.dwDesiredAlphaBPP = 4;
	else
		tsi.dwDesiredAlphaBPP = 0;

	if( dwFlags & DRGTEXTR_32BITSPERPIXEL )
	{
		ddsd.dwAlphaBitDepth = 8;
		tsi.dwDesiredBPP = 32;
		tsi.dwDesiredAlphaBPP = 8;
	}

	
	// Enumerate the texture formats, and find the closest device-supported
	// texture pixel format
    pd3dDevice->EnumTextureFormats( TextureSearchCallback, &tsi );

	//if we tried to create a 32bpp and it didn't work, try 16... (3Dfx, PowerVR PCX2)
	if (( dwFlags & DRGTEXTR_32BITSPERPIXEL ) && (tsi.bFoundGoodFormat == false))
	{
		tsi.dwDesiredBPP = 16;
		if(tsi.dwDesiredAlphaBPP == 8)
			tsi.dwDesiredAlphaBPP = 4;
		dwFlags -= DRGTEXTR_32BITSPERPIXEL;			
		pd3dDevice->EnumTextureFormats( TextureSearchCallback, &tsi );
	}

	// If a palettized format was requested, but not found, default to a
	// 16-bit texture format
	if( FALSE == tsi.bFoundGoodFormat && bUsePalette )
	{
		tsi.bUsePalette  = FALSE;
		tsi.dwDesiredBPP = 16;
	    pd3dDevice->EnumTextureFormats( TextureSearchCallback, &tsi );
		if( FALSE == tsi.bFoundGoodFormat )
			return E_FAIL;
	}
	
	// Create a new surface for the texture
    HRESULT hr;
    if( FAILED( hr = pDD->CreateSurface( &ddsd, &pddsSurface, NULL ) ) )
	{
		OutputDebugString("Couldn't create temp surface\n");
        return E_FAIL;
	}

    // Create the texture
    if( FAILED( pddsSurface->QueryInterface( IID_IDirect3DTexture2,
                                         (VOID**)&ptexTexture ) ) )
        return E_FAIL;


	// Copy the bitmap to the texture surface
	if (hbmAlphaBitmap)
		return CopyBitmapToSurface( pddsSurface, hbmBitmap, hbmAlphaBitmap, dwFlags );
	else
		return CopyBitmapToSurface( pddsSurface, hbmBitmap, NULL, dwFlags );

}




//-----------------------------------------------------------------------------
// Name: CopyBitmapToSurface()
// Desc: Copies the image of a bitmap into a surface
// (should really make this the only function, and the AlphaBitmap name set to NULL
//  by the user means no alpha)
//-----------------------------------------------------------------------------
HRESULT CDRGTexture::CopyBitmapToSurface( LPDIRECTDRAWSURFACE4 pddsTarget,
                                    HBITMAP hbmBitmap,
									HBITMAP hbmAlphaBitmap,									
									DWORD dwFlags )
{
    // Get a DDraw object, we'll need it to create a temporary surface
    LPDIRECTDRAW4 pDD;
    pddsTarget->GetDDInterface( (VOID**)&pDD );
    

    // Get the bitmap structure (to extract width, height, and bpp)
    BITMAP bm;
    GetObject( hbmBitmap, sizeof(BITMAP), &bm );
	
	//do the same for the alpha bitmap, if there is one
	BITMAP bmAlpha;
	if (hbmAlphaBitmap)
	{
		GetObject( hbmAlphaBitmap, sizeof(BITMAP), &bmAlpha );
		if ((bmAlpha.bmWidth != bm.bmWidth) || (bmAlpha.bmHeight != bm.bmHeight))
		{
			OutputDebugString ("color and alpha bitmaps not the same size (this isn't supported yet \n");
			return NULL;
		} //end if not same size
	} //end if Alpha bitmap exists
    
			
    // Setup the new surface desc
    DDSURFACEDESC2 ddsd;
	ZeroMemory( &ddsd, sizeof(DDSURFACEDESC2) );
    ddsd.dwSize                 = sizeof(DDSURFACEDESC2);
    ddsd.dwFlags                = 0;
    ddsd.ddsCaps.dwCaps         = 0;
    ddsd.ddpfPixelFormat.dwSize = sizeof(DDPIXELFORMAT);
	
	// get info on the target surface, so we can create one similar, but in sys mem
    pddsTarget->GetSurfaceDesc( &ddsd );

	if ( (dwFlags & DRGTEXTR_CREATEBUMPMAP_DUDV) &&
		 (ddsd.ddpfPixelFormat.dwFlags & DDPF_BUMPDUDV))
	{
		 OutputDebugString ("found the requested bump map pixel format...\n");
	}
    ddsd.dwFlags          = DDSD_CAPS|DDSD_HEIGHT|DDSD_WIDTH|DDSD_PIXELFORMAT|
		                    DDSD_TEXTURESTAGE;
    ddsd.ddsCaps.dwCaps   = DDSCAPS_TEXTURE|DDSCAPS_SYSTEMMEMORY;
    ddsd.ddsCaps.dwCaps2  = 0L;
    ddsd.dwWidth          = bm.bmWidth;
    ddsd.dwHeight         = bm.bmHeight;

    // Create a new surface for the texture
    LPDIRECTDRAWSURFACE4 pddsTempSurface;
	HRESULT hr;
    if( FAILED( hr = pDD->CreateSurface( &ddsd, &pddsTempSurface, NULL ) ) )
        return NULL;

    // Get a Device Context for the bitmap. We'll need it for blitting
    HDC hdcBitmap = CreateCompatibleDC( NULL );
	if( NULL == hdcBitmap )
    {
        pddsTempSurface->Release();
        return NULL;
    }
    SelectObject( hdcBitmap, hbmBitmap );
	
	// and get a DC for the alpha bitmap
	HDC hdcAlphaBitmap = CreateCompatibleDC( NULL );
	if( NULL == hdcAlphaBitmap )
    {
        pddsTempSurface->Release();
        return NULL;
    }
    SelectObject( hdcAlphaBitmap, hbmAlphaBitmap );
	

    // Handle palettized textures. Need to attach a palette
	// if it is a palettized texture, we won't handle alpha, for now
    if( ddsd.ddpfPixelFormat.dwRGBBitCount == 8 )
    {
        // Create the color table and parse in the palette
        DWORD pe[256];
        WORD  wNumColors = GetDIBColorTable( hdcBitmap, 0, 256,
                                             (RGBQUAD*)pe );
        for( WORD i=0; i<wNumColors; i++ )
        {
            pe[i] = 0xff000000 + RGB( GetBValue(pe[i]), GetGValue(pe[i]),
                                      GetRValue(pe[i]) );

			// Set alpha for transparent pixels
			if( dwFlags & DRGTEXTR_TRANSPARENTBLACK )
            {
                if( (pe[i]&0x00ffffff) == 0x00000000 )
                    pe[i] &= 0x00ffffff;
            }
            else if( dwFlags & DRGTEXTR_TRANSPARENTWHITE )
            {
                if( (pe[i]&0x00ffffff) == 0x00ffffff )
                    pe[i] &= 0x00ffffff;
            }
		}
        // Create & attach a palette with the bitmap's colors
        LPDIRECTDRAWPALETTE  pPalette;
		if( dwFlags & (DRGTEXTR_TRANSPARENTWHITE|DRGTEXTR_TRANSPARENTBLACK) )
            pDD->CreatePalette( DDPCAPS_8BIT|DDPCAPS_ALPHA, (PALETTEENTRY*)pe, &pPalette, NULL );
        else
            pDD->CreatePalette( DDPCAPS_8BIT, (PALETTEENTRY*)pe, &pPalette, NULL );
        pddsTempSurface->SetPalette( pPalette );
        pddsTarget->SetPalette( pPalette );
        SAFE_RELEASE( pPalette );
    }

	//don't need this anymore
	pDD->Release(); 

    // Copy the bitmap image to the surface.
    HDC hdcSurface;
    if( SUCCEEDED( pddsTempSurface->GetDC( &hdcSurface ) ) )
    {
		//this will copy the RGB over, 
        BitBlt( hdcSurface, 0, 0, bm.bmWidth, bm.bmHeight, hdcBitmap, 0, 0,
                SRCCOPY );

		//moved  from bottom
		pddsTempSurface->ReleaseDC( hdcSurface );
		
		//but now I have to copy the alpha
		//
		DDSURFACEDESC2 lpSurfDesc;
		ZeroMemory( &lpSurfDesc, sizeof(DDSURFACEDESC2) );
		lpSurfDesc.dwSize                 = sizeof(DDSURFACEDESC2);
		lpSurfDesc.dwFlags                = 0;
		lpSurfDesc.ddsCaps.dwCaps         = 0;
		lpSurfDesc.ddpfPixelFormat.dwSize = sizeof(DDPIXELFORMAT);
		
	
		//this next section is only for creating an inverse heightmap, for doing embossing
		// where an inverse heightmap is: color= 0.5-color (it assumes the original is from 0 to 0.5)
		if(dwFlags & DRGTEXTR_CREATEINVHEIGHTMAP)
		{
			//lock surface so we can twiddle bits
			while( pddsTempSurface->Lock( NULL, &lpSurfDesc, 0, NULL ) ==
				   DDERR_WASSTILLDRAWING );
			
			//for the inv bump map, we need to make the colors = 0.5 - color
			for( DWORD y=0; y<ddsd.dwHeight; y++ )
			{
				WORD*  p16 =  (WORD*)((BYTE*)lpSurfDesc.lpSurface + y*lpSurfDesc.lPitch);
				DWORD* p32 = (DWORD*)((BYTE*)lpSurfDesc.lpSurface + y*lpSurfDesc.lPitch);
        
				for( DWORD x=0; x<lpSurfDesc.dwWidth; x++ )
				{
					if( lpSurfDesc.ddpfPixelFormat.dwRGBBitCount == 16 )
					{					
						//Not the quickest way to do this, but it works.
						COLORREF tempcolor = GetPixel(  hdcBitmap, x, y);
						//this just truncates, not the greatest color conv. We could round.
						DWORD r = tempcolor & 0x000000ff;
						DWORD g = tempcolor & 0x0000ff00; 
						DWORD b = tempcolor & 0x00ff0000;
						r = 0x00000080 - r;
						g = 0x00008000 - g;
						b = 0x00800000 - b;
						r &= 0x000000f0;
						g &= 0x0000f000;
						b &= 0x00f00000;

						if	( lpSurfDesc.ddpfPixelFormat.dwRGBAlphaBitMask == 0xf000)
						{
							
							//4444							
							r = r << 4;
							g = g >> 8;
							b = b >> 20;
						}
						else
						{
							
							if ( lpSurfDesc.ddpfPixelFormat.dwRGBAlphaBitMask == 0x8000)
							{
								//1555								
								r = r << 4;
								g = g >> 8;
								b = b >> 20;
							}
							else
							{
								if (lpSurfDesc.ddpfPixelFormat.dwRGBAlphaBitMask == 0x0000)
								//565
								r = r << 8;
								g = g >> 5;
								b = b >> 19;
							}
						}
						
						*p16 = (WORD)(r | g | b ) ;						
						
						*p16 &= 0xf000; //get rid of any alpha there
						*p16 |= (WORD)(r | g | b ) ;						


						p16++;
					}
					if( ddsd.ddpfPixelFormat.dwRGBBitCount == 32 )
					{						
						//colorref is in the  0x00bbggrr format
						COLORREF tempcolor = GetPixel(  hdcBitmap, x, y);

						DWORD r = tempcolor & 0x000000ff;
						DWORD g = tempcolor & 0x0000ff00; 
						DWORD b = tempcolor & 0x00ff0000;
						r = 0x00000080 - r;
						g = 0x00008000 - g;
						b = 0x00800000 - b;
						r &= 0x000000ff;
						g &= 0x0000ff00;
						b &= 0x00ff0000;

						r = r << 16;
						//g = g >> 8; //g is in the right place
						b = b >> 16;
						
						
						*p32 &= 0xff000000; //get rid of any alpha there
						*p32 |= (DWORD)(r | g | b ) ;						
						
						
						p32++;
					}
				}
			} //end for

			pddsTempSurface->Unlock( NULL);
		} //end if	invheightmap					

    } //end if getdc
	
//I think the GetDC call isn't working if this is a bump map
	// so I will try a by-hand copy
	//
	if (ddsd.ddpfPixelFormat.dwFlags & DDPF_BUMPDUDV)
	{
		OutputDebugString(" try writing to the bump...\n");

		DDSURFACEDESC2 lpSurfDesc;
		ZeroMemory( &lpSurfDesc, sizeof(DDSURFACEDESC2) );
		lpSurfDesc.dwSize                 = sizeof(DDSURFACEDESC2);
		lpSurfDesc.dwFlags                = 0;
		lpSurfDesc.ddsCaps.dwCaps         = 0;
		lpSurfDesc.ddpfPixelFormat.dwSize = sizeof(DDPIXELFORMAT);
		
		
		while( pddsTempSurface->Lock( NULL, &lpSurfDesc, 0, NULL ) ==
				   DDERR_WASSTILLDRAWING );

		//now that it's locked, let's go through it..
		for( DWORD y=0; y<ddsd.dwHeight; y++ )
		{
			WORD*  p16 =  (WORD*)((BYTE*)lpSurfDesc.lpSurface + y*lpSurfDesc.lPitch);
			DWORD* p32 = (DWORD*)((BYTE*)lpSurfDesc.lpSurface + y*lpSurfDesc.lPitch);
        
			for( DWORD x=0; x<lpSurfDesc.dwWidth; x++ )
			{
				if( lpSurfDesc.ddpfPixelFormat.dwRGBBitCount == 16 )
					{					
						//Not the quickest way to do this, but it works.
						COLORREF tempcolor = GetPixel(  hdcBitmap, x, y);
						
						DWORD Du = tempcolor & 0x000000ff;
						DWORD Dv = tempcolor & 0x0000ff00; 
										
						
						*p16 = (WORD)(Du | Dv);
						
						//*p16 &= 0xf000; //get rid of any alpha there
						//*p16 |= (WORD)(r | g | b ) ;						


						p16++;
						//p16++;
					}
					if( ddsd.ddpfPixelFormat.dwRGBBitCount == 32 )
					{						
						//colorref is in the  0x00bbggrr format
						COLORREF tempcolor = GetPixel(  hdcBitmap, x, y);

						DWORD Du = tempcolor & 0x0000ffff;
						DWORD Dv = tempcolor & 0xffff0000; 
						
						*p32 |= (DWORD)(Du | Dv) ;						
												
						p32++;
					}
			}
			} //end for
			pddsTempSurface->Unlock( NULL );
		//pddsTarget->Unlock( NULL );

	}
	

    // Copy the temp surface to the real texture surface
    pddsTarget->Blt( NULL, pddsTempSurface, NULL, DDBLT_WAIT, NULL );

    if( ddsd.ddpfPixelFormat.dwRGBBitCount == 8 )
	{
        LPDIRECTDRAWPALETTE  pPalette;
		DWORD pe[256];
        pddsTempSurface->GetPalette( &pPalette );
		pPalette->GetEntries( 0, 0, 256, (PALETTEENTRY*)&pe );
		pPalette->Release();

		pddsTarget->GetPalette( &pPalette );
		pPalette->GetEntries( 0, 0, 256, (PALETTEENTRY*)&pe );
		pPalette->Release();
	}

    pddsTempSurface->Release();

	// For textures with real alpha (not palettized), set transparent bits
	// (for now, if it is a bump map, I don't do anything, but I should enable 
	// the capability to load a separate bmp file for luminance one day...
    if( ddsd.ddpfPixelFormat.dwRGBAlphaBitMask )
	{
		if( dwFlags & (DRGTEXTR_TRANSPARENTWHITE|DRGTEXTR_TRANSPARENTBLACK) )
		{
			// Lock the texture surface
			DDSURFACEDESC2 ddsd;
			ZeroMemory( &ddsd, sizeof(DDSURFACEDESC2) );
			ddsd.dwSize                 = sizeof(DDSURFACEDESC2);
			ddsd.dwFlags                = 0;
			ddsd.ddsCaps.dwCaps         = 0;
			ddsd.ddpfPixelFormat.dwSize = sizeof(DDPIXELFORMAT);

			while( pddsTarget->Lock( NULL, &ddsd, 0, NULL ) ==
				   DDERR_WASSTILLDRAWING );
			 
		    DWORD dwAlphaMask = ddsd.ddpfPixelFormat.dwRGBAlphaBitMask;
			DWORD dwRGBMask   = ( ddsd.ddpfPixelFormat.dwRBitMask |
							      ddsd.ddpfPixelFormat.dwGBitMask |
							      ddsd.ddpfPixelFormat.dwBBitMask );
			DWORD dwColorkey  = 0x00000000; // Colorkey on black
			if( dwFlags & DRGTEXTR_TRANSPARENTWHITE ) 
				dwColorkey = dwRGBMask;     // Colorkey on white

			// Add an opaque alpha value to each non-colorkeyed pixel
			for( DWORD y=0; y<ddsd.dwHeight; y++ )
			{
				WORD*  p16 =  (WORD*)((BYTE*)ddsd.lpSurface + y*ddsd.lPitch);
				DWORD* p32 = (DWORD*)((BYTE*)ddsd.lpSurface + y*ddsd.lPitch);
        
				for( DWORD x=0; x<ddsd.dwWidth; x++ )
				{
					if( ddsd.ddpfPixelFormat.dwRGBBitCount == 16 )
					{
						if( ( *p16 &= dwRGBMask ) != dwColorkey )
							*p16 |= dwAlphaMask;
						p16++;
					}
					if( ddsd.ddpfPixelFormat.dwRGBBitCount == 32 )
					{
						if( ( *p32 &= dwRGBMask ) != dwColorkey )
							*p32 |= dwAlphaMask;
						p32++;
					}
				}
			} //end for
			pddsTarget->Unlock( NULL );
		} //end one bit alpha
		else //end if transparent black/white
		{
			//user asked for alpha, so we need to copy over alpha bits
			//but not if this is a 1555 ARGB texture (only if we have 4 or 8 bits of alpha)

			if ( ddsd.ddpfPixelFormat.dwRGBAlphaBitMask != 0x8000)
			{
			// Lock the texture surface
			DDSURFACEDESC2 ddsd;
			ZeroMemory( &ddsd, sizeof(DDSURFACEDESC2) );
			ddsd.dwSize                 = sizeof(DDSURFACEDESC2);
			ddsd.dwFlags                = 0;
			ddsd.ddsCaps.dwCaps         = 0;
			ddsd.ddpfPixelFormat.dwSize = sizeof(DDPIXELFORMAT);

			while( pddsTarget->Lock( NULL, &ddsd, 0, NULL ) ==
				   DDERR_WASSTILLDRAWING );
			 
		    DWORD dwAlphaMask = ddsd.ddpfPixelFormat.dwRGBAlphaBitMask;
			DWORD dwRGBMask   = ( ddsd.ddpfPixelFormat.dwRBitMask |
							      ddsd.ddpfPixelFormat.dwGBitMask |
							      ddsd.ddpfPixelFormat.dwBBitMask );
			DWORD dwColorkey  = 0x00000000; // Colorkey on black

			// Add an opaque alpha value to each non-colorkeyed pixel
			for( DWORD y=0; y<ddsd.dwHeight; y++ )
			{
				WORD*  p16 =  (WORD*)((BYTE*)ddsd.lpSurface + y*ddsd.lPitch);
				DWORD* p32 = (DWORD*)((BYTE*)ddsd.lpSurface + y*ddsd.lPitch);
        
				for( DWORD x=0; x<ddsd.dwWidth; x++ )
				{
					if( ddsd.ddpfPixelFormat.dwRGBBitCount == 16 )
					{
						
						COLORREF tempcolor = GetPixel(  hdcBitmap, x, y);
						//this just truncates, not the greatest color conv.
						DWORD r = tempcolor & 0x000000f0; //keep 4 msb
						DWORD g = tempcolor & 0x0000f000; 
						DWORD b = tempcolor & 0x00f00000;
						r = r << 4;
						g = g >> 8;
						b = b >> 20;

						DWORD r_fraction = tempcolor & 0x00000008; //keep 4 msb
						DWORD g_fraction = tempcolor & 0x00000800; 
						DWORD b_fraction = tempcolor & 0x00080000;
						r_fraction = r_fraction << 5;
						g_fraction = g_fraction >> 7;
						b_fraction = b_fraction >> 19;


						*p16 = (WORD)(r | g | b ) ;						
						
						//colorref is in the  0x00bbggrr format
						tempcolor = GetPixel(  hdcAlphaBitmap, x, y);
						tempcolor &= 0x000000f0; //keep 4 bits of red channel
						tempcolor = tempcolor << 8; //only 8, because it's the most sig. 4 bits, 
						*p16 &= 0x0fff; //get rid of any alpha there
						*p16 |= (WORD)tempcolor; //add alpha from our alpha bitmap


						p16++;
					}
					if( ddsd.ddpfPixelFormat.dwRGBBitCount == 32 )
					{
						
						//colorref is in the  0x00bbggrr format
						COLORREF tempcolor = GetPixel(  hdcAlphaBitmap, x, y);
						tempcolor &= 0x000000ff; //keep 8 bits of red channel
						tempcolor = tempcolor << 24;
						*p32 = *p32 & 0x00ffffff; //get rid of any alpha there
						*p32 |= tempcolor; //add alpha from our alpha bitmap

						p32++;
					}
				}
			}
			
			pddsTarget->Unlock( NULL );

			} //end if not 1555				
		} //end else
	} //end if

    DeleteDC( hdcBitmap );
	DeleteDC( hdcAlphaBitmap );

    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: D3DUtil_GetDirectDrawFromDevice()
// Desc: Get the DDraw interface from a D3DDevice.
//-----------------------------------------------------------------------------
LPDIRECTDRAW4 CDRGTexture::GetDirectDrawFromDevice( LPDIRECT3DDEVICE3 pd3dDevice )
{
	LPDIRECTDRAW4        pDD = NULL;
	LPDIRECTDRAWSURFACE4 pddsRender;

    if( pd3dDevice )
	{
	    // Get the current render target
		if( SUCCEEDED( pd3dDevice->GetRenderTarget( &pddsRender ) ) )
		{
		    // Get the DDraw4 interface from the render target
			pddsRender->GetDDInterface( (VOID**)&pDD );
			pddsRender->Release();
		}
	}
	return pDD;
}


