// DRGWrapper.cpp
// ------------------------------------------------------------------------------------
// Copyright  1999 Intel Corporation
// All Rights Reserved
// 
// Permission is granted to use, copy, distribute and prepare derivative works of this 
// software for any purpose and without fee, provided, that the above copyright notice
// and this statement appear in all copies.  Intel makes no representations about the
// suitability of this software for any purpose.  This software is provided "AS IS." 
//
// Intel specifically disclaims all warranties, express or implied, and all liability,
// including consequential and other indirect damages, for the use of this software, 
// including liability for infringement of any proprietary rights, and including the 
// warranties of merchantability and fitness for a particular purpose.  Intel does not
// assume any responsibility for any errors which may appear in this software nor any
// responsibility to update it.
// ------------------------------------------------------------------------------------
//
//  PURPOSE:
//    
// DRGWrapper.cpp: implementation of the CDRGWrapper class.
// This class handles all device & mode enumeration, surface creation, viewport setup, etc.
//
// ------------------------------------------------------------------------------------
//
// Authors: Dean Macri, Kim Pallister - Intel Developer Relations & Engineering Group
//
////////////////////////////////////////////////////////////////////////////////////////

#include "DRGShell.h"

//
// The following three file globals are used during the enumeration
// process because we're combining the DDraw+D3D driver pairs into
// one driver for each pair.
//
static GUID g_guidDDraw;	
static GUID *g_pguidDDraw;
static LPSTR g_pDDrawDescription;
static LPSTR g_pDDrawName;
static LPDIRECTDRAW4	g_pDirectDraw4;

BOOL WINAPI EnumDDrawDeviceCallback( GUID FAR *lpGUID, 
											LPSTR lpDriverDescription,
											LPSTR lpDriverName,
											LPVOID lpContext, 
											HMONITOR hm );
HRESULT CALLBACK EnumD3DDevicesCallback(  GUID FAR* lpGuid,                    
												 LPSTR lpDeviceDescription,             
												 LPSTR lpDeviceName,                  
												 LPD3DDEVICEDESC lpD3DHWDeviceDesc,     
												 LPD3DDEVICEDESC lpD3DHELDeviceDesc,  
												 LPVOID lpContext );
HRESULT WINAPI EnumModesCallback(  LPDDSURFACEDESC2 lpDDSurfaceDesc,  
										  LPVOID lpContext );
HRESULT WINAPI EnumZBufferFormatsCallback( DDPIXELFORMAT* pddpf, void* pddpfDesired );

// -----------------------------------------------------------------------
// CDRGWrapper::CDRGWrapper( HWND hWnd)
//
// -----------------------------------------------------------------------
CDRGWrapper::CDRGWrapper( HWND hWnd )
{
	m_iDrivers = 0;
	m_iCurrentDriver = -1;
	m_iCurrentMode = -1;
	m_pFirstDriver = NULL;
	m_pCurrentDriver = NULL;
	m_pCurrentMode = NULL;
	m_bInitialized = false;
	m_bFullscreen = false;
	m_bHardware = false;
	m_bZBuffer = false;
	m_dwBackBufferCount = 0;
	m_dwFSRenderWidth = 640;
	m_dwFSRenderHeight = 480;
	m_dwFSBpp = 16;

	m_fFrameTime = 0.0f;

	m_pDD = NULL;
	m_pD3D = NULL;
	m_pFrontBuffer = NULL;
	m_pBackBuffer = NULL;
	m_pZBuffer = NULL;
	m_pD3DDevice = NULL;
	m_pD3DViewport = NULL;

	m_hWnd = hWnd;
	//
	// Enumerate the available DirectDraw devices and in the process
	// enumerate the available D3D Devices for each DirectDraw device
	//
	HRESULT hr = DirectDrawEnumerateEx( EnumDDrawDeviceCallback, this, 
						DDENUM_ATTACHEDSECONDARYDEVICES |
						DDENUM_DETACHEDSECONDARYDEVICES |
						DDENUM_NONDISPLAYDEVICES );
	//
	// Should throw an exception here if we couldn't get any devices
	//
	if( FAILED( hr ) )
	{
	}
}

// -----------------------------------------------------------------------
// CDRGWrapper::~CDRGWrapper()
//
// -----------------------------------------------------------------------
CDRGWrapper::~CDRGWrapper()
{
	if( m_iCurrentDriver >= 0 )
		DestroyMode();

	//
	// Delete the drivers
	//
	D3DDriver *pDriver = m_pFirstDriver;
	D3DDriver *pDriverTemp = pDriver;
	while( pDriver )
	{
		//
		// Delete the modes for the driver
		//
		D3DFSMode *pMode = pDriver->pFSModes;
		D3DFSMode *pModeTemp = pMode;
		while( pMode )
		{
			pMode = pMode->pNext;
			delete pModeTemp;
			pModeTemp = pMode;
		}

		SAFE_DELETE( pDriver->pDDrawDescription );
		SAFE_DELETE( pDriver->pDDrawName );
		SAFE_DELETE( pDriver->pD3DDescription );
		SAFE_DELETE( pDriver->pD3DName );
		SAFE_DELETE( pDriver->pMenuDescription );

		pDriver = pDriver->pNext;
		delete pDriverTemp;
		pDriverTemp = pDriver;
	}

}

// -----------------------------------------------------------------------
// D3DDriver *CDRGWrapper::GetDriver( int iDriver )
//
// -----------------------------------------------------------------------
D3DDriver *CDRGWrapper::GetDriver( int iDriver )
{
	if( (iDriver < 0) || (iDriver >= m_iDrivers) )
		return NULL;

	D3DDriver *pDriver = m_pFirstDriver;

	while( iDriver )
	{
		iDriver--;
		pDriver = pDriver->pNext;
	}
	return pDriver;
}

// -----------------------------------------------------------------------
// D3DFSMode *CDRGWrapper::GetMode( int iDriver, int iMode )
//
// -----------------------------------------------------------------------
D3DFSMode *CDRGWrapper::GetMode( int iDriver, int iMode )
{
	D3DDriver *pDriver = GetDriver( iDriver );
	if( !pDriver )
		return NULL;

	//
	// Check if we have any modes for this driver -- if not, enumerate them
	//
	if( !pDriver->iNumModes )
	{
		EnumModes( iDriver, NULL, NULL );
	}

	if( (iMode < 0) || (iMode >= pDriver->iNumModes ) )
		return NULL;

	D3DFSMode *pMode = pDriver->pFSModes;
	while( iMode )
	{
		iMode--;
		pMode = pMode->pNext;
	}

	return pMode;
}


// -----------------------------------------------------------------------
// int CDRGWrapper::GetNumDriverss( void )
//
// -----------------------------------------------------------------------
int CDRGWrapper::GetNumDrivers( void )
{
	return m_iDrivers;
}

// -----------------------------------------------------------------------
// int CDRGWrapper::GetNumModes( int iDriver )
//
// -----------------------------------------------------------------------
int CDRGWrapper::GetNumModes( int iDriver )
{
	D3DDriver *pDriver = GetDriver( iDriver );

	if( pDriver )
	{
		if( !pDriver->iNumModes )
		{
			//
			// Enumerate the modes
			//
			EnumModes( iDriver, NULL, NULL );
		}
		return pDriver->iNumModes;
	}

	return 0;
}

// -----------------------------------------------------------------------
// HRESULT CDRGWrapper::GetBestDriver( int &iDriver, bool bHardware, bool bAGP, DWORD dwOtherFlags )
//
// -----------------------------------------------------------------------
HRESULT CDRGWrapper::GetBestDriver( int &iDriver, bool bHardware, bool bAGP, DWORD dwOtherFlags )
{
	D3DDriver *pDriver = m_pFirstDriver;
	int iCurrentDriver = 0;

	iDriver = -1;

	bool bDriverGood = false;

	while( pDriver )
	{
		if( bHardware )
		{
			if( pDriver->HALDesc.dwFlags )
			{
				if( bAGP )
				{
					if( pDriver->HALDesc.dwDevCaps & D3DDEVCAPS_TEXTURENONLOCALVIDMEM )
					{
						//
						// Check dwOtherFlags here
						//
						//iDriver = iCurrentDriver;
						//return S_OK;
						bDriverGood = true;
					}
				}
				else
				{
					//
					// Check dwOtherFlags here
					//
					//iDriver = iCurrentDriver;
					//return S_OK;
					bDriverGood = true;
				}
			}
		}
		else
		{
			//
			// Check dwOtherFlags here
			//

			//
			// For now, if you don't ask for hardware, you get the first
			// driver in the list
			//
			//iDriver = iCurrentDriver;
			//return S_OK;
			bDriverGood = true;
		}

		//if the driver supports the caps bits requested so far, we need to check the other 
		//required ones
		if (bDriverGood)
		{
			//note: I am using the separate texture memories flag, as this is a sure sign that
			// the HW cannot render to a texture surface (you could get around it by blitting,
			// but I haven't done this. Some drivers are improperly succeeding to a call to
			// CreateSurface when the D3DDevice and D3DTexture flags are both specified. The
			// HW should fail thew call to create the surface, but does not.

			//HW - bump - target
			if ( (bHardware) && (dwOtherFlags & REQUIREBUMPMAPPING) && 
				                (dwOtherFlags & REQUIRETARGETTEXTURE) )
			{
				if (((pDriver->HALDesc.dwDevCaps & D3DDEVCAPS_SEPARATETEXTUREMEMORIES) == 0) &&
					(pDriver->HALDesc.dwTextureOpCaps & D3DTEXOPCAPS_BUMPENVMAP))
				{					
					iDriver = iCurrentDriver;
					return S_OK;			
				}				
			}
			else
			{
				// HW - no bump - target
				if ((bHardware) && (dwOtherFlags & REQUIRETARGETTEXTURE))
				{
					if ((bHardware) && 
						((pDriver->HALDesc.dwDevCaps & D3DDEVCAPS_SEPARATETEXTUREMEMORIES) == 0))
					{
						iDriver = iCurrentDriver;
						return S_OK;			
					}				
				}
				// HW - bump - no target
				if ((bHardware) && (dwOtherFlags & REQUIREBUMPMAPPING))
				{
					if ((bHardware) && 
						(pDriver->HALDesc.dwTextureOpCaps & D3DTEXOPCAPS_BUMPENVMAP))
					{
						iDriver = iCurrentDriver;
						return S_OK;			
					}				
				}		

			}

			//SW - bump (no need to check for target texture support, SW devices support them)
			if ( (!bHardware) && (dwOtherFlags & REQUIREBUMPMAPPING) )
			{
				if (pDriver->HELDesc.dwTextureOpCaps & D3DTEXOPCAPS_BUMPENVMAP)					
				{					
					iDriver = iCurrentDriver;
					return S_OK;			
				}
			}

			//if we have gotten to here, it means we did not find the device we wanted, so we will fall through to
			// the next device on the list			
		}
		
		pDriver = pDriver->pNext;
		iCurrentDriver++;
	}

	return D3DERR_INVALID_DEVICE;
}

// -----------------------------------------------------------------------
// HRESULT CDRGWrapper::EnumModes( int iDriver, int *iModes, D3DFSMode **pModes )
//
// -----------------------------------------------------------------------
HRESULT CDRGWrapper::EnumModes( int iDriver, int *iModes, D3DFSMode **pModes )
{
	D3DDriver *pDriver = GetDriver( iDriver );
	D3DDriver *pHoldDriver = m_pCurrentDriver;
	LPDIRECTDRAW pDD = NULL;
	LPDIRECTDRAW4 pDD4 = NULL;
	HRESULT hr;

	if( !pDriver )
		return D3DERR_INVALID_DEVICE;


	if( iDriver != m_iCurrentDriver )
	{
		hr = DirectDrawCreate( pDriver->pguidDDraw, &pDD, NULL );
		HANDLE_ERROR( hr, hr );

		//
		// Query for IDirectDraw4
		//
		hr = pDD->QueryInterface( IID_IDirectDraw4, (void **)&pDD4 );
		SAFE_RELEASE( pDD );
		HANDLE_ERROR( hr, hr );
	}
	else
	{
		pDD4 = m_pDD;
	}

	m_pCurrentDriver = pDriver;

	//
	// Delete any existing modes
	//
	D3DFSMode *pMode = pDriver->pFSModes;
	D3DFSMode *pModeTemp = pMode;
	while( pMode )
	{
		pMode = pMode->pNext;
		delete pModeTemp;
		pModeTemp = pMode;
	}
	pDriver->iNumModes = 0;

	hr = pDD4->EnumDisplayModes( 0, NULL, this, EnumModesCallback );
	m_pCurrentDriver = pHoldDriver;

	HANDLE_ERROR( hr, hr );

	if( iDriver != m_iCurrentDriver )
		SAFE_RELEASE( pDD4 );

	if( pModes )
		*pModes = pDriver->pFSModes;
	if( iModes )
		*iModes = pDriver->iNumModes;

	return S_OK;
}

// -----------------------------------------------------------------------
// HRESULT CDRGWrapper::InitMode( int iDriver, int iMode, DWORD dwBackBufferCount, bool bZBuffer )
//
// -----------------------------------------------------------------------
HRESULT CDRGWrapper::InitMode( int iDriver, int iMode, DWORD dwBackBufferCount, bool bZBuffer )
{
	m_pCurrentDriver = GetDriver( iDriver );
	m_pCurrentMode = GetMode( iDriver, iMode );

	return InitMode( iDriver, true, m_pCurrentMode->dwWidth, m_pCurrentMode->dwHeight,
			  m_pCurrentMode->dwBpp, dwBackBufferCount, bZBuffer, false );
}

// -----------------------------------------------------------------------
// HRESULT CDRGWrapper::InitMode( int iDriver, bool bFullScreen,
//							  DWORD dwWidth, DWORD dwHeight, DWORD dwBpp, DWORD dwBackBufferCount,
//							  bool bZBuffer, bool bUseDefaultOnError )
//
// -----------------------------------------------------------------------
HRESULT CDRGWrapper::InitMode( int iDriver, bool bFullScreen, 
							   DWORD dwWidth, DWORD dwHeight, DWORD dwBpp, DWORD dwBackBufferCount,
							   bool bZBuffer, bool bUseDefaultOnError )
{
	DWORD dwFlags = 0;
	HRESULT hr;

	//
	// If we've already initialized a mode, shut it down
	//
	if( m_iCurrentDriver >= 0 )
		DestroyMode();

	//
	// Do some parameter checking
	//
	if( dwBackBufferCount < 1 )
		return ERROR_INVALID_PARAMETER;

	m_pCurrentDriver = GetDriver( iDriver );
	if( !m_pCurrentDriver )
		return ERROR_INVALID_PARAMETER;

	//
	// Save the window position
	//
	GetWindowRect( m_hWnd, &m_rcWindow );


	m_bHardware = (m_pCurrentDriver->HALDesc.dwFlags != 0 );

	m_bFullscreen = (bFullScreen || m_pCurrentDriver->bFullScreenOnly); // FS if requested or mandatory
	m_dwBpp = dwBpp;
	if( !m_bFullscreen )
	{
		RECT rcView;

		GetClientRect( m_hWnd, &rcView );
		m_dwRenderWidth = rcView.right;
		m_dwRenderHeight = rcView.bottom;
	}
	else
	{
		m_dwRenderWidth = dwWidth;
		m_dwRenderHeight = dwHeight;
		//
		// Keep these around for toggling between fullscreen and non-fullscreen
		//
		m_dwFSRenderWidth = dwWidth;
		m_dwFSRenderHeight = dwHeight;
		m_dwFSBpp = dwBpp;
	}
	m_bZBuffer = bZBuffer;
	m_dwBackBufferCount = dwBackBufferCount;

	hr = CreateDirectDraw();
	ON_ERROR_GOTO( hr, InitCleanup );

	hr = CreateD3D();
	ON_ERROR_GOTO( hr, InitCleanup );

	hr = CreateSurfaces( dwBackBufferCount );
	ON_ERROR_GOTO( hr, InitCleanup );

	//
	// Create and attach the z-buffer if requested
	//
	if( bZBuffer )
	{
		hr = CreateZBuffer();
		ON_ERROR_GOTO( hr, InitCleanup );
	}

	//
	// Create the 3D device
	//
	if( FAILED( hr = Create3DDevice() ) )
		return hr;

	//
	// Create and set the 3D viewport
	//
	if( FAILED( hr = CreateViewport() ) )
		return hr;

	m_iCurrentDriver = iDriver;

	m_bInitialized = true;

	return S_OK;

InitCleanup:
	DestroyMode();

	return hr;
}

// -----------------------------------------------------------------------
// HRESULT CDRGWrapper::DestroyMode( void )
//
// -----------------------------------------------------------------------
HRESULT CDRGWrapper::DestroyMode( void )
{
	HRESULT hr;

	if( !m_bInitialized )
		return S_OK;

	m_bInitialized = false;

	SAFE_RELEASE( m_pD3DViewport );
	SAFE_RELEASE( m_pD3DDevice );

	SAFE_RELEASE( m_pZBuffer );
	SAFE_RELEASE( m_pBackBuffer );
	SAFE_RELEASE( m_pFrontBuffer );
	SAFE_RELEASE( m_pD3D );

	if( m_pDD )
	{
		char szTemp[256];
		hr = m_pDD->SetCooperativeLevel( m_hWnd, DDSCL_NORMAL );
		if( FAILED( hr ) )
		{
			sprintf( szTemp, "SetCooperativeLevel failed: %8lx\r\n", hr );
			OutputDebugString( szTemp );
		}
	}
	SAFE_RELEASE( m_pDD );
	if( m_bFullscreen )
		SetWindowPos( m_hWnd, HWND_NOTOPMOST, m_rcWindow.left, m_rcWindow.top,
							m_rcWindow.right - m_rcWindow.left,
							m_rcWindow.bottom - m_rcWindow.top,
							SWP_SHOWWINDOW );


	m_pCurrentDriver = NULL;
	m_pCurrentMode = NULL;
	m_iCurrentDriver = -1;
	m_iCurrentMode = -1;
	m_dwRenderWidth = 0;
	m_dwRenderHeight = 0;
	m_bFullscreen = false;

	return S_OK;
}

// -----------------------------------------------------------------------
// HRESULT CDRGWrapper::Clear( DWORD dwFlags )
//
// -----------------------------------------------------------------------
HRESULT CDRGWrapper::Clear( DWORD dwFlags )
{
	if( !m_bInitialized )
		return S_OK;

	DWORD dwClearColor = 0;
	D3DVALUE dvZClearValue = 1.0f;
	DWORD dwStencilClearValue = 0;
	D3DRECT rcClear;
	HRESULT hr;

	rcClear.x1 = 0;
	rcClear.y1 = 0;
	rcClear.x2 = m_dwRenderWidth;
	rcClear.y2 = m_dwRenderHeight;

	hr = m_pD3DViewport->Clear2( 1, &rcClear, dwFlags,
							dwClearColor, dvZClearValue, dwStencilClearValue );

	return hr;
}

// -----------------------------------------------------------------------
// HRESULT CDRGWrapper::StartFrame( void )
//
// -----------------------------------------------------------------------
HRESULT CDRGWrapper::StartFrame( void )
{
	if( !m_bInitialized )
		return S_OK;

	return m_pD3DDevice->BeginScene();
}

// -----------------------------------------------------------------------
// HRESULT CDRGWrapper::EndFrame( void )
//
// -----------------------------------------------------------------------
HRESULT CDRGWrapper::EndFrame( void )
{
	if( !m_bInitialized )
		return S_OK;

	HRESULT hr;

	hr = m_pD3DDevice->EndScene();
	HANDLE_ERROR( hr, hr );

	if( m_bFullscreen )
	{
		//
		// Full screen mode, just flip
		//
		hr = m_pFrontBuffer->Flip( NULL, DDFLIP_WAIT );
		HANDLE_ERROR( hr, hr );
	}
	else
	{
		//
		// Windowed mode, do a Blt
		//
		hr = m_pFrontBuffer->Blt( &m_rcScreen, m_pBackBuffer, 
								  &m_rcView, DDBLT_WAIT, NULL );
		HANDLE_ERROR( hr, hr );
	}

	return S_OK;
}

// -----------------------------------------------------------------------
// HRESULT CDRGWrapper::MoveWindow( void )
//
// -----------------------------------------------------------------------
HRESULT CDRGWrapper::MoveWindow( void )
{
	if( !m_bInitialized )
		return S_OK;

	if( !m_bFullscreen )
	{
		GetWindowRect( m_hWnd, &m_rcWindow );

		GetClientRect( m_hWnd, &m_rcView );
		GetClientRect( m_hWnd, &m_rcScreen );
		ClientToScreen( m_hWnd, (POINT*)&m_rcScreen.left );
		ClientToScreen( m_hWnd, (POINT*)&m_rcScreen.right );
	}

	return S_OK;
}

// -----------------------------------------------------------------------
// HRESULT CDRGWrapper::RefreshMode( bool bFullscreen )
//
// -----------------------------------------------------------------------
HRESULT CDRGWrapper::RefreshMode( bool bFullscreen )
{
	if( !m_bInitialized )
		return S_OK;

	int iDriver = m_iCurrentDriver;
	DWORD dwRenderWidth = m_dwRenderWidth;
	DWORD dwRenderHeight = m_dwRenderHeight;
	DWORD dwBpp = m_dwBpp;
	DWORD dwBackBufferCount = m_dwBackBufferCount;
	bool bZBuffer = m_bZBuffer;
	HRESULT hr;

	if( bFullscreen )
	{
		dwRenderWidth = m_dwFSRenderWidth;
		dwRenderHeight = m_dwFSRenderHeight;
		dwBpp = m_dwFSBpp;
	}

	//
	// The mode has changed (typically the result of a window resizing)
	// Destroy the current mode and re-initialize it.
	//
	hr = DestroyMode();
	HANDLE_ERROR( hr, hr );

	hr = InitMode( iDriver, bFullscreen, dwRenderWidth,
				   dwRenderHeight, dwBpp, dwBackBufferCount,
				   bZBuffer, false );
	HANDLE_ERROR( hr, hr );

	return S_OK;
}

// -----------------------------------------------------------------------
// HRESULT CDRGWrapper::ChangeDriver( int iDriver )
//
// -----------------------------------------------------------------------
HRESULT CDRGWrapper::ChangeDriver( int iDriver )
{
	if( !m_bInitialized )
		return S_OK;

	DWORD dwRenderWidth = m_dwRenderWidth;
	DWORD dwRenderHeight = m_dwRenderHeight;
	DWORD dwBpp = m_dwBpp;
	DWORD dwBackBufferCount = m_dwBackBufferCount;
	bool bZBuffer = m_bZBuffer;
	bool bFullscreen = m_bFullscreen;
	HRESULT hr;

	//need to check the driver being selected, to see if it is FS only
	D3DDriver	*pDrv = m_pFirstDriver;
	int iDrvNum = 0;
		
	while ((iDrvNum++ != iDriver) && (pDrv->pNext !=NULL))
		pDrv = pDrv->pNext;
	//we should now be pointed at the driver that the user selected
	if (pDrv->bFullScreenOnly)
	{
		m_bFullscreen = true;
		bFullscreen = true;
	}

	if( bFullscreen )
	{
		dwRenderWidth = m_dwFSRenderWidth;
		dwRenderHeight = m_dwFSRenderHeight;
		dwBpp = m_dwFSBpp;
	}

	//
	// Destroy the current mode and re-initialize it with the new driver
	//
	hr = DestroyMode();
	HANDLE_ERROR( hr, hr );

	hr = InitMode( iDriver, bFullscreen, dwRenderWidth,
				   dwRenderHeight, dwBpp, dwBackBufferCount,
				   bZBuffer, false );
	HANDLE_ERROR( hr, hr );

	return S_OK;
}


// -----------------------------------------------------------------------
// HRESULT CALLBACK EnumD3DDevicesCallback(  GUID FAR* lpGuid,
//												 LPSTR lpDeviceDescription,             
//												 LPSTR lpDeviceName,                  
//												 LPD3DDEVICEDESC lpD3DHWDeviceDesc,     
//												 LPD3DDEVICEDESC lpD3DHELDeviceDesc,  
//												 LPVOID lpContext )
//
// -----------------------------------------------------------------------
HRESULT CALLBACK EnumD3DDevicesCallback(  GUID FAR* lpGuid,                    
												 LPSTR lpDeviceDescription,             
												 LPSTR lpDeviceName,                  
												 LPD3DDEVICEDESC lpD3DHWDeviceDesc,     
												 LPD3DDEVICEDESC lpD3DHELDeviceDesc,  
												 LPVOID lpContext )
{
    CDRGWrapper *pWrapper = (CDRGWrapper *)lpContext;

	//
	// Validate the parameters we received
	//
	if( (lpGuid==NULL) || (lpD3DHWDeviceDesc==NULL) ||
		(lpD3DHELDeviceDesc== NULL) ||
		(lpDeviceDescription==NULL) || (lpDeviceName==NULL) )
		return DDENUMRET_OK;

	//
	// Handle some special devices
	//
	if( IsEqualGUID( *lpGuid, IID_IDirect3DNullDevice ) )
		return DDENUMRET_OK;

	//
	// Allocate memory for the device and fill in the information
	//
	D3DDriver *pNewDriver;

	if( !pWrapper->m_pFirstDriver )
	{
		pWrapper->m_pFirstDriver = new D3DDriver;
		pNewDriver = pWrapper->m_pFirstDriver;
	}
	else
	{
		D3DDriver *pLastDriver = pWrapper->m_pFirstDriver;

		while( pLastDriver->pNext )
			pLastDriver = pLastDriver->pNext;
		pLastDriver->pNext = new D3DDriver;
		pNewDriver = pLastDriver->pNext;
	}

	pNewDriver->pNext = NULL;
	pNewDriver->pFSModes = NULL;
	pNewDriver->iNumModes = 0;

	if( g_pguidDDraw )
	{
		memcpy( &pNewDriver->guidDDrawDevice, &g_guidDDraw, sizeof( GUID ) );
		pNewDriver->pguidDDraw = &pNewDriver->guidDDrawDevice;
	}
	else
		pNewDriver->pguidDDraw = NULL;

	memcpy( &pNewDriver->guidD3DDevice, lpGuid, sizeof( GUID ) );
	pNewDriver->pguidD3D = &pNewDriver->guidD3DDevice;
	if( g_pDDrawDescription )
	{
		pNewDriver->pDDrawDescription = new char[lstrlen( g_pDDrawDescription )+1];
		lstrcpy( pNewDriver->pDDrawDescription, g_pDDrawDescription );
	}
	else
		pNewDriver->pDDrawDescription = NULL;
	if( g_pDDrawName )
	{
		pNewDriver->pDDrawName = new char[lstrlen( g_pDDrawName )+1];
		lstrcpy( pNewDriver->pDDrawName, g_pDDrawName );
	}
	else
		pNewDriver->pDDrawName = NULL;

	pNewDriver->pD3DDescription = new char[lstrlen( lpDeviceDescription )+1];
	lstrcpy( pNewDriver->pD3DDescription, lpDeviceDescription );
	pNewDriver->pD3DName = new char[lstrlen( lpDeviceName )+1];
	lstrcpy( pNewDriver->pD3DName, lpDeviceName );
	pNewDriver->pMenuDescription = new char[ lstrlen( g_pDDrawDescription ) + lstrlen( lpDeviceName ) + 4 ];
	wsprintf( pNewDriver->pMenuDescription, "%s - %s", g_pDDrawDescription, lpDeviceName );

	memcpy( &pNewDriver->HALDesc, lpD3DHWDeviceDesc, sizeof( D3DDEVICEDESC ) );
	memcpy( &pNewDriver->HELDesc, lpD3DHELDeviceDesc, sizeof( D3DDEVICEDESC ) );

	//This is where I will try setting the flags for 3Dfx-type cards that must run FS. 
	
	//create a  ddcaps struct, helcapsstruct, blank'em out
	DDCAPS	ddDriverCaps;
	DDCAPS	ddDriverHELCaps;
	ddDriverCaps.dwSize = sizeof(DDCAPS);
	ddDriverHELCaps.dwSize = sizeof(DDCAPS);
	
	g_pDirectDraw4->GetCaps(&ddDriverCaps,&ddDriverHELCaps); 
	//check the caps for the flag
	if (ddDriverCaps.dwCaps2 & DDCAPS2_CANRENDERWINDOWED)
		pNewDriver->bFullScreenOnly = false;
	else
		pNewDriver->bFullScreenOnly = true;


	pWrapper->m_iDrivers++;

	return DDENUMRET_OK;
}

// -----------------------------------------------------------------------
// BOOL WINAPI EnumDDrawDeviceCallback( GUID FAR *lpGUID, 
// 									 LPSTR lpDriverDescription,
//									 LPSTR lpDriverName,
//									 LPVOID lpContext, 
//									 HMONITOR hm )
//
// Callback for enumerating DirectDraw devices.  Each DirectDraw
// device could have multiple D3D Devices.
// -----------------------------------------------------------------------
BOOL WINAPI EnumDDrawDeviceCallback( GUID FAR *lpGUID, 
									 LPSTR lpDriverDescription,
									 LPSTR lpDriverName,
									 LPVOID lpContext, 
									 HMONITOR hm )
{
    CDRGWrapper *pWrapper = (CDRGWrapper *)lpContext;
	LPDIRECTDRAW pDD = NULL;
	LPDIRECTDRAW4 pDD4 = NULL;
	LPDIRECT3D3 pD3D = NULL;
	HRESULT hr;

	g_pguidDDraw = lpGUID;
	if( g_pguidDDraw )
		memcpy( &g_guidDDraw, lpGUID, sizeof( GUID ) );
	g_pDDrawDescription = lpDriverDescription;
	g_pDDrawName = lpDriverName;

	hr = DirectDrawCreate( lpGUID, &pDD, NULL );
	HANDLE_ERROR( hr, TRUE );

	
	//
	// Query for IDirectDraw4
	//
	hr = pDD->QueryInterface( IID_IDirectDraw4, (void **)&pDD4 );
	SAFE_RELEASE( pDD );
	HANDLE_ERROR( hr, TRUE );

	//
	// Query for access to Direct3D
	//
	hr = pDD4->QueryInterface( IID_IDirect3D3, (void **)&pD3D );
	if( FAILED( hr ) )
	{
		SAFE_RELEASE( pDD4 );
		return TRUE;
	}
	g_pDirectDraw4 = pDD4; //point the global over so that we can use it in the d3ddevice enum

	//
	// Enumerate the various D3D devices (Hardware, Software, etc.) for this 
	// DirectDraw device
	//
	pD3D->EnumDevices( EnumD3DDevicesCallback, lpContext );

	SAFE_RELEASE( pD3D );
	SAFE_RELEASE( pDD4 );

	return TRUE;
}

// -----------------------------------------------------------------------
// HRESULT WINAPI EnumModesCallback(  LPDDSURFACEDESC2 lpDDSurfaceDesc,
//									  LPVOID lpContext )
//
// -----------------------------------------------------------------------
HRESULT WINAPI EnumModesCallback(  LPDDSURFACEDESC2 lpDDSurfaceDesc,  
								   LPVOID lpContext )
{
    CDRGWrapper *pWrapper = (CDRGWrapper *)lpContext;
	D3DFSMode *pNewMode;

	if( !lpDDSurfaceDesc )
		return DDENUMRET_OK;
	//
	// Ignore modes that are not 16 or 32 bit
	//
	if(( lpDDSurfaceDesc->ddpfPixelFormat.dwRGBBitCount != 16 ) &&
		( lpDDSurfaceDesc->ddpfPixelFormat.dwRGBBitCount != 32 ) )
		return DDENUMRET_OK;

    if( pWrapper->m_pCurrentDriver->pFSModes )
	{
        D3DFSMode *pTemp = pWrapper->m_pCurrentDriver->pFSModes;
		while( pTemp->pNext )
			pTemp = pTemp->pNext;
		pTemp->pNext = new D3DFSMode;
		pNewMode = pTemp->pNext;
	}
	else
	{
        pNewMode = pWrapper->m_pCurrentDriver->pFSModes = new D3DFSMode;
	}

	memcpy( &pNewMode->SurfDesc, lpDDSurfaceDesc, sizeof( DDSURFACEDESC2 ) );
	pNewMode->dwWidth = lpDDSurfaceDesc->dwWidth;
	pNewMode->dwHeight = lpDDSurfaceDesc->dwHeight;
	pNewMode->dwBpp = lpDDSurfaceDesc->ddpfPixelFormat.dwRGBBitCount;
	wsprintf( pNewMode->szDescription, "%d x %d - %d Bit", 
		pNewMode->dwWidth,
		pNewMode->dwHeight,
		pNewMode->dwBpp );
	pNewMode->pNext = NULL;
	pWrapper->m_pCurrentDriver->iNumModes++;

	return DDENUMRET_OK;
}



// -----------------------------------------------------------------------
// HRESULT WINAPI EnumZBufferFormatsCallback( DDPIXELFORMAT* pddpf, void* pddpfDesired )
//
// -----------------------------------------------------------------------
HRESULT WINAPI EnumZBufferFormatsCallback( DDPIXELFORMAT* pddpf, void* pddpfDesired )
{
    if( !pddpf || !pddpfDesired )
        return D3DENUMRET_CANCEL;

	//
	// Check if the supported pixel format matches the desired format
	//
    if( pddpf->dwFlags == ((DDPIXELFORMAT*)pddpfDesired)->dwFlags )
    {
        memcpy( pddpfDesired, pddpf, sizeof(DDPIXELFORMAT) );

		//
		// We'll be happy with 16-bit or higher
		//
		if( pddpf->dwZBufferBitDepth >= 16 )
			return D3DENUMRET_CANCEL;
    }

    return D3DENUMRET_OK;
}



// -----------------------------------------------------------------------
// HRESULT CDRGWrapper::CreateDirectDraw( void )
//
// -----------------------------------------------------------------------
HRESULT CDRGWrapper::CreateDirectDraw( void )
{
	LPDIRECTDRAW pDD = NULL;
	DWORD dwFlags;
	HRESULT hr;

	//
	// Create the IDirectDraw4 interface. DirectDrawCreate only creates a DDraw interface
	// so we need to query for the DDraw4 interface.
	//
	hr = DirectDrawCreate( m_pCurrentDriver->pguidDDraw, &pDD, NULL );
	HANDLE_ERROR( hr, hr );
	hr = pDD->QueryInterface( IID_IDirectDraw4, (void **)&m_pDD );
	SAFE_RELEASE( pDD );
	HANDLE_ERROR( hr, hr );

	//
	// Set the DirectDraw cooperative level
	//
	if( m_bFullscreen )
		dwFlags = DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN; // | DDSCL_FPUSETUP;
	else
		dwFlags = DDSCL_NORMAL; // | DDSCL_FPUSETUP;

	hr = m_pDD->SetCooperativeLevel( m_hWnd, dwFlags );
	HANDLE_ERROR( hr, hr );

	return S_OK;
}

// -----------------------------------------------------------------------
// HRESULT CDRGWrapper::CreateD3D( void )
//
// -----------------------------------------------------------------------
HRESULT CDRGWrapper::CreateD3D( void )
{
	HRESULT hr;

	//
	// Create the IDirect3D3 interface.
	//
	hr = m_pDD->QueryInterface( IID_IDirect3D3, (void **)&m_pD3D );
	HANDLE_ERROR( hr, hr );

	//
	// Switch the mode if necessary
	//
	if( m_bFullscreen )
	{
		DWORD dwModeFlags = 0;

		hr = m_pDD->SetDisplayMode( m_dwRenderWidth, m_dwRenderHeight,
									m_dwBpp, 0, dwModeFlags );
		HANDLE_ERROR( hr, hr );
	}

	return S_OK;
}

// -----------------------------------------------------------------------
// HRESULT CDRGWrapper::CreateSurfaces( DWORD dwBackBufferCount )
//
// -----------------------------------------------------------------------
HRESULT CDRGWrapper::CreateSurfaces( DWORD dwBackBufferCount )
{
	HRESULT hr;
	DDSURFACEDESC2 ddsd;
	DDSCAPS2 ddscaps;

	//
	// Create the primary surface
	//
	memset( &ddsd, 0, sizeof( DDSURFACEDESC2 ) );
	ddsd.dwSize = sizeof( DDSURFACEDESC2 );
	ddsd.ddpfPixelFormat.dwSize = sizeof( DDPIXELFORMAT );
	if( m_bFullscreen )
	{
		ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_3DDEVICE | DDSCAPS_FLIP | DDSCAPS_COMPLEX;
		ddsd.dwBackBufferCount = dwBackBufferCount;
	}
	else
	{
		ddsd.dwFlags = DDSD_CAPS;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
	}

	hr = m_pDD->CreateSurface( &ddsd, &m_pFrontBuffer, NULL );
	HANDLE_ERROR( hr, hr );

	if( !m_bFullscreen )
	{
		//
		// Create a clipper if we're rendering to a window
		//
		LPDIRECTDRAWCLIPPER pClipper;
		hr = m_pDD->CreateClipper( 0, &pClipper, NULL );
		HANDLE_ERROR( hr, hr );

		//
		// Attach the clipper to the front buffer
		//
		pClipper->SetHWnd( 0, m_hWnd );
		m_pFrontBuffer->SetClipper( pClipper );
		SAFE_RELEASE( pClipper );
	}


	//
	// Create the back buffers
	//
	if( m_bFullscreen )
	{
		ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
		hr = m_pFrontBuffer->GetAttachedSurface( &ddscaps, &m_pBackBuffer );
		HANDLE_ERROR( hr, hr );

		//
		// Setup the rectangles for blting
		//
		SetRect( &m_rcView, 0, 0, m_dwRenderWidth, m_dwRenderHeight );
		memcpy( &m_rcScreen, &m_rcView, sizeof(RECT) );
	}
	else
	{
		ddsd.dwFlags        = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS;
		ddsd.dwWidth        = m_dwRenderWidth;
		ddsd.dwHeight       = m_dwRenderHeight;
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_3DDEVICE;

		hr = m_pDD->CreateSurface( &ddsd, &m_pBackBuffer, NULL );
		HANDLE_ERROR( hr, hr );

		//
		// Setup the blt rectangles
		//
		GetClientRect( m_hWnd, &m_rcView );
		GetClientRect( m_hWnd, &m_rcScreen );
		ClientToScreen( m_hWnd, (POINT*)&m_rcScreen.left );
		ClientToScreen( m_hWnd, (POINT*)&m_rcScreen.right );
	}

	return S_OK;
}

// -----------------------------------------------------------------------
// HRESULT CDRGWrapper::CreateZBuffer( void )
//
// -----------------------------------------------------------------------
HRESULT CDRGWrapper::CreateZBuffer( void )
{
	DDSURFACEDESC2 ddsd;
	DDPIXELFORMAT ddpfZBuffer;
	HRESULT hr;

	//
	// Enumerate the supported z-buffer formats to find one we like
	//
	memset( &ddpfZBuffer, 0, sizeof(DDPIXELFORMAT) );
	ddpfZBuffer.dwFlags = DDPF_ZBUFFER;	// EFTU: Can add DDPF_STENCILBUFFER to 
										//       get a stencil buffer

    hr = m_pD3D->EnumZBufferFormats( m_pCurrentDriver->guidD3DDevice, EnumZBufferFormatsCallback,
                                (void*)&ddpfZBuffer );
	HANDLE_ERROR( hr, hr );

    if( sizeof(DDPIXELFORMAT) != ddpfZBuffer.dwSize )
        return D3DERR_ZBUFFER_NOTPRESENT;

	//
	// Request a z-buffer in the appropriate format
	//
	memset( &ddsd, 0, sizeof( DDSURFACEDESC2 ) );
	ddsd.dwSize = sizeof( DDSURFACEDESC2 );
	ddsd.dwFlags = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS | DDSD_PIXELFORMAT;
	if( m_bHardware )
		ddsd.ddsCaps.dwCaps = DDSCAPS_ZBUFFER | DDSCAPS_VIDEOMEMORY;
	else
		ddsd.ddsCaps.dwCaps = DDSCAPS_ZBUFFER | DDSCAPS_SYSTEMMEMORY;
	ddsd.dwWidth = m_dwRenderWidth;
	ddsd.dwHeight = m_dwRenderHeight;
	memcpy( &ddsd.ddpfPixelFormat, &ddpfZBuffer, sizeof( DDPIXELFORMAT ) );

	//
	// EFTU:  Could check for the D3DPRASTERCAPS_ZBUFFERLESSHSR capability
	//		  and not allocate a z-buffer if the bit is set
	//

	//
	// Create the z-buffer
	//
	hr = m_pDD->CreateSurface( &ddsd, &m_pZBuffer, NULL );
	HANDLE_ERROR( hr, hr );

	//
	// Attach the z-buffer to the back buffer (ideally you would attach the
	// z-buffer to the render target but the way we've implemented things here,
	// the back buffer is always the render target).
	// 
	hr = m_pBackBuffer->AddAttachedSurface( m_pZBuffer );
	HANDLE_ERROR( hr, hr );

    return S_OK;

}

// -----------------------------------------------------------------------
// HRESULT CDRGWrapper::Create3DDevice( void )
//
// -----------------------------------------------------------------------
HRESULT CDRGWrapper::Create3DDevice( void )
{
	DDSURFACEDESC2 ddsd;
	HRESULT hr;

	//
	// We aren't going to support palettes, so make sure we're not
	// in a palettized mode.  EFTU:  Support palettes if you want to (Ha Ha)
	//
	ddsd.dwSize = sizeof(DDSURFACEDESC2);
	m_pDD->GetDisplayMode( &ddsd );
	if( ddsd.ddpfPixelFormat.dwRGBBitCount <= 8 )
		return D3DERR_INVALID_DEVICE;

	//
	// Create the device
	//
	hr = m_pD3D->CreateDevice( m_pCurrentDriver->guidD3DDevice, m_pBackBuffer,
								&m_pD3DDevice, NULL );
	HANDLE_ERROR( hr, hr );

    return S_OK;

}

// -----------------------------------------------------------------------
// HRESULT CDRGWrapper::CreateViewport( void )
//
// -----------------------------------------------------------------------
HRESULT CDRGWrapper::CreateViewport( void )
{
    HRESULT      hr;
    D3DVIEWPORT2 viewData;

	//
	// Create the viewport
	//
	hr = m_pD3D->CreateViewport( &m_pD3DViewport, NULL );
	HANDLE_ERROR( hr, hr );

	//
	// Attach the view to the 3D device
	//
	hr = m_pD3DDevice->AddViewport( m_pD3DViewport );
	HANDLE_ERROR( hr, hr );


	//
	// Initialize the parameters for the viewport
	//
	memset( &viewData, 0, sizeof( D3DVIEWPORT2 ) );

	viewData.dwSize = sizeof( D3DVIEWPORT2 );
	viewData.dwWidth = m_dwRenderWidth;
	viewData.dwHeight = m_dwRenderHeight;

	viewData.dvMaxZ = 1.0f;
	viewData.dvClipX = -1.0f;
	viewData.dvClipY = 1.0f;
	viewData.dvClipWidth = 2.0f;
	viewData.dvClipHeight = 2.0f;

	hr = m_pD3DViewport->SetViewport2( &viewData );
	HANDLE_ERROR( hr, hr );

	//
	// Set this viewport as the current viewport for the 3D device
	//
	hr = m_pD3DDevice->SetCurrentViewport(  m_pD3DViewport );
	HANDLE_ERROR( hr, hr );

    return S_OK;
}


