// DRGWrapper.cpp
// ------------------------------------------------------------------------------------
// Copyright  1999 Intel Corporation
// All Rights Reserved
// 
// Permission is granted to use, copy, distribute and prepare derivative works of this 
// software for any purpose and without fee, provided, that the above copyright notice
// and this statement appear in all copies.  Intel makes no representations about the
// suitability of this software for any purpose.  This software is provided "AS IS." 
//
// Intel specifically disclaims all warranties, express or implied, and all liability,
// including consequential and other indirect damages, for the use of this software, 
// including liability for infringement of any proprietary rights, and including the 
// warranties of merchantability and fitness for a particular purpose.  Intel does not
// assume any responsibility for any errors which may appear in this software nor any
// responsibility to update it.
// ------------------------------------------------------------------------------------
//
//  PURPOSE:
//    
// DRGWrapper.cpp: implementation of the CDRGWrapper class.
// This class handles all device & mode enumeration, surface creation, viewport setup, etc.
//
// ------------------------------------------------------------------------------------
//
// Authors: Dean Macri, Kim Pallister - Intel Developer Relations & Engineering Group
//
////////////////////////////////////////////////////////////////////////////////////////

#if !defined(AFX_DRGWRAPPER_H__CFF24424_47D6_11D2_B1A5_00A0C90D2C1B__INCLUDED_)
#define AFX_DRGWRAPPER_H__CFF24424_47D6_11D2_B1A5_00A0C90D2C1B__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#define		REQUIREBUMPMAPPING		0x00000001	//flags added for this demo
#define		REQUIRETARGETTEXTURE	0x00000002	


//Structure defining an individual display mode. A linked list of
//these is constructed from the selected device.
typedef struct tagD3DFSMode
{
	DWORD dwWidth;				// Width (in pixels, of course)	
	DWORD dwHeight;				// Height
	DWORD dwBpp;				// Color depth
	char szDescription[30];		// 
	DDSURFACEDESC2 SurfDesc;	// The DirectDraw Surface Descriptor describing the surface.
								//  the width and height are contained in the descriptor, but 
								//  have been placed in this structure.
	struct tagD3DFSMode *pNext;	// ptr to next mode in the linked list.
} D3DFSMode;

//Structure describing a Device found. A linked list of these is constructed during
// device enumeration
typedef struct tagD3DDriver
{
	GUID guidDDrawDevice;	// The system could have multiple DirectDraw devices
	GUID guidD3DDevice;		// Each DDraw device usually has multiple d3d Devices (HAL,HEL)	
	GUID *pguidDDraw;		// We need pointers to the GUIDs because a NULL
	GUID *pguidD3D;			// GUID is a valid GUID, but we also need to copy
							// the GUID data from the data based to the enumeration
							// callbacks by DirectX.  Hence, pguidDDraw will either
							// be NULL or point to guidDDrawDevice, and likewise
							// for pguidD3D.

	LPSTR pDDrawDescription;
	LPSTR pDDrawName;
	LPSTR pD3DDescription;
	LPSTR pD3DName;
	LPSTR pMenuDescription;	// A string that will be used in our menu to list the device.
	D3DDEVICEDESC HALDesc;  // Device descriptors for the Hardware and emulation devices.
	D3DDEVICEDESC HELDesc;
	int iNumModes;			// The number of modes enumerated.
	D3DFSMode *pFSModes;    // pointer to the linked list of modes for this device.

	bool bFullScreenOnly;	// special just for 3Dfx-type cards
	
	struct tagD3DDriver *pNext; //ptr to the linked list of devices.
} D3DDriver;

// The Wrapper class itself. 
class CDRGWrapper  
{
private:
	int m_iDrivers;				//Total number of devices that were enumerated
	int m_iCurrentDriver;		//The current one being used.
	int m_iCurrentMode;			//The current mode on that device being used.
	bool m_bInitialized;		
	bool m_bZBuffer;			// Is a Z buffer being used.
	DWORD m_dwBackBufferCount;	// How many back buffers are being used? (Not always one!)
	DWORD m_dwBpp;				// Current Color depth
	DWORD m_dwFSRenderWidth;	// Current resolution
	DWORD m_dwFSRenderHeight;	//
	DWORD m_dwFSBpp;			// Color depth in Full Screen (remember, we may toggle)
	RECT m_rcWindow;			// Location of the window in windowed mode

	RECT m_rcScreen, m_rcView;	//

	D3DDriver *m_pFirstDriver;	// pointer to the beginning of the Device linked list.
	//D3DDriver *m_pCurrentDriver;// pointer to the device being used.	
	D3DFSMode *m_pCurrentMode;	// pointer to the current mode being used on that device.


	//
	// Helper functions for creating the stuff required to initialize a mode
	//
	HRESULT CreateDirectDraw( void );
	HRESULT CreateD3D( void );
	HRESULT CreateSurfaces( DWORD dwBackBufferCount );
	HRESULT CreateZBuffer( void );
	HRESULT Create3DDevice( void );
	HRESULT CreateViewport( void );

	//
	// The following member variables are public for now, but they should
	// probably be replaced with private variables and public access methods
	//
public:

	//These two were changed from private to public since the app may need to check them. i.e. 
	//If I do an init mode, and then wonder if what I got was hardware...
	bool m_bFullscreen;			// We toggle between FS and windowed. This tracks it.
	bool m_bHardware;			// Is hardware rendering being used.

	HWND m_hWnd;				// Window handle
	DWORD m_dwRenderWidth;
	DWORD m_dwRenderHeight;
	
	LPDIRECTDRAW4 m_pDD;
	LPDIRECT3D3 m_pD3D;
	LPDIRECTDRAWSURFACE4 m_pFrontBuffer;
	LPDIRECTDRAWSURFACE4 m_pBackBuffer;
	LPDIRECTDRAWSURFACE4 m_pZBuffer;
	LPDIRECT3DDEVICE3 m_pD3DDevice;
	LPDIRECT3DVIEWPORT3 m_pD3DViewport;
	D3DDriver *m_pCurrentDriver;// pointer to the device being used.	

	float m_fFrameTime;			//current frame time.
	
public:
    CDRGWrapper( HWND hWnd );
    virtual ~CDRGWrapper();

	//
	// Internal functions for dealing with the linked lists of drivers and modes
	//
	virtual D3DDriver *GetDriver( int iDriver );
	virtual D3DFSMode *GetMode( int iDriver, int iMode );

	virtual bool Initialized( void ) { return m_bInitialized; }

	virtual int GetNumDrivers( void );
	virtual int GetNumModes( int iDriver );

	
	virtual HRESULT GetBestDriver( int &iDriver,				//Helper function that lets the app
								   bool bHardware = true,		// get the best driver for the job, 
								   bool bAGP = false,			// based on some common requests 
								   DWORD dwOtherFlags = 0 );    // (hardware? AGP? etc)
	virtual HRESULT EnumModes( int iDriver, int *iModes, D3DFSMode **pModes );
	
	virtual HRESULT InitMode( int iDriver,						//Helper function to init the
							  int iMode,						//chosen driver to the chose mode.
							  DWORD dwBackBufferCount = 1,		//Init mode is overloaded so you
							  bool bZBuffer = true );			//can call it with either a mode
	virtual HRESULT InitMode( int iDriver = 0,					//number (look for the one you want)
							  bool bFullScreen = true,			//or specify some parameters and
							  DWORD dwWidth = 640,				//have it look for the right mode.
							  DWORD dwHeight = 480, 
							  DWORD dwBpp = 16,
							  DWORD dwBackBufferCount = 1,
							  bool bZBuffer = true, 
							  bool bUseDefaultOnError = true );
	virtual HRESULT DestroyMode( void );
	virtual HRESULT Clear( DWORD dwFlags );
	virtual HRESULT StartFrame( void );
	virtual HRESULT EndFrame( void );

	virtual HRESULT MoveWindow( void );
	virtual HRESULT RefreshMode( bool bFullscreen );
	virtual HRESULT ChangeDriver( int iDriver );

	friend BOOL WINAPI EnumDDrawDeviceCallback( GUID FAR *lpGUID, 
												LPSTR lpDriverDescription,
												LPSTR lpDriverName,
												LPVOID lpContext, 
												HMONITOR hm );
	friend HRESULT CALLBACK EnumD3DDevicesCallback(  GUID FAR* lpGuid,                    
													 LPSTR lpDeviceDescription,             
													 LPSTR lpDeviceName,                  
													 LPD3DDEVICEDESC lpD3DHWDeviceDesc,     
													 LPD3DDEVICEDESC lpD3DHELDeviceDesc,  
													 LPVOID lpContext );
	friend HRESULT WINAPI EnumModesCallback(  LPDDSURFACEDESC2 lpDDSurfaceDesc,  
											  LPVOID lpContext );

};

#endif // !defined(AFX_DRGWRAPPER_H__CFF24424_47D6_11D2_B1A5_00A0C90D2C1B__INCLUDED_)
